<?php

namespace Drupal\sirv\Plugin\Field\FieldFormatter;

use Drupal\sirv\SirvUrlInterface;
use Drupal\image\Plugin\Field\FieldFormatter\ImageFormatter;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'sirv_image' formatter.
 *
 * @FieldFormatter(
 *   id = "sirv_image",
 *   label = @Translation("Sirv image"),
 *   field_types = {
 *     "image"
 *   },
 *   weight = 10
 * )
 */
class SirvImageFormatter extends ImageFormatter {

  /**
   * The Sirv settings.
   *
   * @var array
   */
  protected $sirvSettings;

  /**
   * The Sirv URL service.
   *
   * @var \Drupal\sirv\SirvUrlInterface
   */
  protected $sirvUrlService;

  /**
   * Constructs a SirvImageFormatter object.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings settings.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\Core\Entity\EntityStorageInterface $image_style_storage
   *   The image style storage.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\sirv\SirvUrlInterface $sirv_url_service
   *   The Sirv URL service.
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, AccountInterface $current_user, EntityStorageInterface $image_style_storage, ConfigFactoryInterface $config_factory, SirvUrlInterface $sirv_url_service) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings, $current_user, $image_style_storage);

    $this->sirvSettings = $config_factory->get('sirv.settings')->get();
    $this->sirvUrlService = $sirv_url_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('current_user'),
      $container->get('entity.manager')->getStorage('image_style'),
      $container->get('config.factory'),
      $container->get('sirv.url')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'sirv_profile' => '',
      'extra_sirv_options' => '',
      'responsive' => TRUE,
      'responsive_scale_crop' => '',
      'responsive_resize_step' => '',
      'lazy_load' => TRUE,
      'lazy_load_threshold' => '',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    // Get the parent's settings form.
    $element = parent::settingsForm($form, $form_state);

    // Customize the description for the Image Style setting.
    $element['image_style']['#description'] = $this->t('The original image is usually desired, though an image style may be selected if it provides an effect that is not available in Sirv, such as custom cropping.');

    $element['sirv_profile'] = [
      '#title' => $this->t('Sirv profile'),
      '#type' => 'textfield',
      '#description' => $this->t('Enter a Sirv profile to use for processing images in this field.'),
      '#default_value' => $this->getSetting('sirv_profile'),
    ];
    $element['extra_sirv_options'] = [
      '#title' => $this->t('Extra Sirv options'),
      '#type' => 'textfield',
      '#description' => $this->t('These options will override options from the profile. Example of format: %example', ['%example' => 'scale.width=250&grayscale=true']),
      '#default_value' => $this->getSetting('extra_sirv_options'),
    ];

    $element['responsive'] = [
      '#title' => $this->t('Make responsive'),
      '#type' => 'checkbox',
      '#description' => $this->t('Check to enable responsive behavior for this field.'),
      '#default_value' => $this->getSetting('responsive'),
    ];
    $element['responsive_scale_crop'] = [
      '#title' => $this->t('Automatic scale/crop'),
      '#type' => 'select',
      '#description' => $this->t('Select an option for scaling and cropping an image to fit a container.'),
      '#default_value' => $this->getSetting('responsive_scale_crop'),
      '#options' => [
        'contain' => $this->t('Contain (default)'),
        'cover' => $this->t('Cover'),
        'crop' => $this->t('Crop'),
      ],
    ];
    $element['responsive_resize_step'] = [
      '#title' => $this->t('Resize step'),
      '#type' => 'number',
      '#description' => $this->t('Enter the change in browser width required for a new responsive image to be generated.'),
      '#default_value' => $this->getSetting('responsive_resize_step'),
      '#size' => 10,
      '#field_suffix' => $this->t('px'),
    ];

    $element['lazy_load'] = [
      '#title' => $this->t('Use lazy loading'),
      '#type' => 'checkbox',
      '#description' => $this->t('Check to enable lazy loading for this field.'),
      '#default_value' => $this->getSetting('lazy_load'),
    ];
    $element['lazy_load_threshold'] = [
      '#title' => $this->t('Threshold'),
      '#type' => 'textfield',
      '#description' => $this->t('Enter a threshold value to indicate when to lazy load the image before it comes into view.'),
      '#default_value' => $this->getSetting('lazy_load_threshold'),
      '#size' => 10,
    ];

    // Move the Image Link setting to the bottom of the form.
    $element['image_link']['#weight'] = 10;

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    // Get the parent's settings summary.
    $summary = parent::settingsSummary();

    if (!empty($this->getSetting('sirv_profile'))) {
      $summary[] = $this->t('Sirv profile: @profile', ['@profile' => $this->getSetting('sirv_profile')]);
    }

    if (!empty($this->getSetting('extra_sirv_options'))) {
      $summary[] = $this->t('Extra Sirv options');
    }

    if (!empty($this->getSetting('responsive'))) {
      $summary[] = $this->t('Responsive');
    }

    if (!empty($this->getSetting('lazy_load'))) {
      $summary[] = $this->t('Lazy load');
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    // Get the parent's elements.
    $elements = parent::viewElements($items, $langcode);

    $responsive = $this->getSetting('responsive');
    $lazy_load = $this->getSetting('lazy_load');

    foreach ($elements as &$element) {
      $element = $this->processElement($element);
    }

    if ($responsive || $lazy_load) {
      // Add the Sirv library.
      $elements['#attached']['library'] = 'sirv/sirv';

      // Add header links to preconnect and prefetch DNS.
      $elements['#attached']['html_head_link'][][] = [
        'rel' => 'preconnect',
        'href' => 'https://scripts.sirv.com',
        'crossorigin' => TRUE,
      ];
      $elements['#attached']['html_head_link'][][] = [
        'rel' => 'dns-prefetch',
        'href' => 'https://scripts.sirv.com',
      ];
    }

    // Add header links to preconnect and prefetch DNS.
    $elements['#attached']['html_head_link'][][] = [
      'rel' => 'preconnect',
      'href' => 'https://' . $this->sirvSettings['domain'],
      'crossorigin' => TRUE,
    ];
    $elements['#attached']['html_head_link'][][] = [
      'rel' => 'dns-prefetch',
      'href' => 'https://' . $this->sirvSettings['domain'],
    ];

    return $elements;
  }

  /**
   * Process an element to add Sirv-specific data.
   *
   * @param array $element
   *   The element to be processed.
   *
   * @return array
   *   The processed element.
   */
  protected function processElement(array $element) {
    $field_settings = $this->getSettings();
    $responsive = $this->getSetting('responsive');
    $lazy_load = $this->getSetting('lazy_load');

    $data_options = [];

    if (!isset($element['#item'])) {
      return $element;
    }

    $item = $element['#item'];
    if (!$file = $item->entity) {
      return $element;
    }

    // Get the Sirv URL.
    $sirv_url = $this->sirvUrlService->generateFromFile($file, $field_settings);
    if (!$sirv_url) {
      return $element;
    }

    // Make sure the "class" attribute exists.
    if (!isset($element['#item_attributes']['class'])) {
      $element['#item_attributes']['class'] = [];
    }

    if ($responsive || $lazy_load) {
      // Set the URI to a placeholder value so it can be removed in
      // preprocessing.
      $element['#item']->set('uri', SirvUrlInterface::NULL);

      $element['#item_attributes']['data-src'] = $sirv_url;
      $element['#item_attributes']['class'][] = 'Sirv';
    }
    else {
      $element['#item']->set('uri', $sirv_url);
    }

    if ($responsive) {
      if (!$lazy_load) {
        $data_options['lazy'] = 'false';
      }

      // Handle the Responsive Scale/Crop setting.
      switch ($field_settings['responsive_scale_crop']) {
        case 'contain':
          $element['#item_attributes']['class'][] = 'image-fit';
          break;

        case 'cover':
          $element['#item_attributes']['class'][] = 'image-fill';
          $data_options['fit'] = 'cover';
          break;

        case 'crop':
          $element['#item_attributes']['class'][] = 'image-fill';
          $data_options['fit'] = 'crop';
          break;
      }

      // Handle the Responsive Resize Step setting.
      if ($field_settings['responsive_resize_step']) {
        $data_options['resizeStep'] = $field_settings['responsive_resize_step'];
      }

      // Remove any assigned width or height.
      $value = $element['#item']->getValue();
      unset($value['width']);
      unset($value['height']);
      $element['#item']->setValue($value);
    }

    if ($lazy_load) {
      if (!$responsive) {
        $data_options['resize'] = 'false';
      }

      // Handle the Lazy Load Threshold setting.
      if ($field_settings['lazy_load_threshold']) {
        $data_options['threshold'] = $field_settings['lazy_load_threshold'];
      }
    }

    $data_options_attribute = NULL;
    if (!empty($data_options)) {
      $data_options_items = [];
      foreach ($data_options as $key => $value) {
        $data_options_items[] = "$key:$value";
      }
      $data_options_attribute = implode(';', $data_options_items);
    }

    if ($data_options_attribute) {
      $element['#item_attributes']['data-options'] = $data_options_attribute;
    }

    return $element;
  }

}
